/*
  Syntax:

  #const
  const
  Y:const
  Y:X
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define IMAGELEN (1<<13)
#define DX_EXTENSION 1

unsigned char objectCode[IMAGELEN];

int errorCnt, pass2, lineNr, curAddr[5], segment, here, yCont, yOpt, listing = 3, endFlag, dx_ext, z_ext;

#define NSYMBOLS 1050 // works best if somewhat bigger than power of 2
#define SYMLEN 20
#define DEFINED   (1<<31)
#define MNEMONIC  (1<<30)
#define DIRECTIVE (1<<29)

#define SAVA   0xFF // address in zero page for saving A during video generation
#define CYCLES 0xFE // cycle counter in zero page
#define VIDEO  0x00 // offset in code page of video handler

int key[NSYMBOLS];
int value[NSYMBOLS];
int tabSize, hashMask;

#define M_IMMEDIATE    (1<<8)
#define M_ZEROPAGE     (1<<9)
#define M_INDEXED      (1<<10)
#define M_ABSOLUTE     (1<<11)
#define M_YINDEXED     (1<<12)
#define M_ACCUMULATOR  (1<<13)
#define M_INPUT        (1<<14)
#define M_ACCUTARGET   (1<<15)
#define M_STORE        (1<<16)
#define M_OUTPUT       (1<<17)
#define M_INCREMENT    (1<<18)
#define M_INTARGET     (1<<19)
#define M_INDIRECT     (1<<20)
#define M_YPREFIX      (1<<21)
#define M_NOMODE       (1<<22)
#define M_OPTIONAL     (1<<23)
#define M_REGISTER     (M_ACCUMULATOR | M_INPUT | M_IMMEDIATE)
#define M_MEMORY       (M_ZEROPAGE | M_INDEXED | M_ABSOLUTE | M_YINDEXED | M_INCREMENT)
#define M_ALU          (M_REGISTER | M_MEMORY | M_OUTPUT)
#define M_BRA          (M_ZEROPAGE | M_ACCUTARGET | M_INTARGET | M_INDIRECT)
#define M_MODE         (M_ALU | M_BRA)


typedef struct {
  char *name;
  int opcode;
} Mnemonic;

Mnemonic mnemonics[] = {
 {"lda",   0 + M_ALU },
 {"and",  32 + M_ALU },
 {"ora",  64 + M_ALU },
 {"xor",  96 + M_ALU },
 {"add", 128 + M_ALU },
 {"sub", 160 + M_ALU },
 {"sta", 194 + M_MEMORY + M_STORE },
 {"stin",195 + M_MEMORY + M_STORE },
 {"mov", 192 + M_IMMEDIATE + M_STORE },
 {"jmp", 224 + M_BRA + M_YPREFIX },
 {"bgt", 228 + M_BRA },
 {"blt", 232 + M_BRA },
 {"bne", 236 + M_BRA },
 {"beq", 240 + M_BRA },
 {"bge", 244 + M_BRA },
 {"ble", 248 + M_BRA },
 {"bra", 252 + M_BRA },
 {"ldx",  16 + M_ACCUMULATOR + M_IMMEDIATE + M_ZEROPAGE + M_INPUT },
 {"ldy",  20 + M_ACCUMULATOR + M_IMMEDIATE + M_ZEROPAGE + M_INPUT },
 {"ldp",  20 + M_ACCUMULATOR + M_IMMEDIATE + M_ZEROPAGE + M_INPUT + M_OPTIONAL },
 {"out",  24 + M_ACCUMULATOR + M_IMMEDIATE + M_ZEROPAGE + M_INPUT + M_INCREMENT},
 {"tax", 208 + M_ACCUMULATOR + M_IMMEDIATE + M_ZEROPAGE + M_INPUT },
 {"tay", 212 + M_ACCUMULATOR + M_IMMEDIATE + M_ZEROPAGE + M_INPUT },
 {"nop", 0x02 },
 {"clr", 0x62 },
 {"lsl", 0x82 },
 {"end",  DIRECTIVE + 0 },
 {"text", DIRECTIVE + 1 },
 {"data", DIRECTIVE + 2 },
 {"zero", DIRECTIVE + 7 },
 {"fast", DIRECTIVE + 8 },
 {"vert", DIRECTIVE + 9 },
 {"page", DIRECTIVE + 3 },
 {"byte", DIRECTIVE + 4 },
 {"area", DIRECTIVE + 5 },
 {"rep",  DIRECTIVE + 6 },
 {"sync", DIRECTIVE + 11 },
 {"sysa", DIRECTIVE + 10 },
 {NULL}
};

int official=1;

int Expression (char **p);

int White(char **p)
{
  int commentType = 0;
  while(**p == ' ' || **p == '\t') (*p)++;
  if(**p == '/') {
    if((*p)[1] == '/') commentType = 3; else
    if((*p)[1] == '*') commentType = 4;
  } else if(**p == ';') commentType = 3;
  if(commentType) while(**p && **p != '\n') (*p)++;
  return commentType & ~4;
}

int EndOfLine(char **p)
{ // check if end of line or comment is reached, and skip the latter
  int OK;
  White(p);
  OK = (**p == '\n');
  while(**p && **p != '\n') (*p)++;
  return OK;
}

char symBuf[SYMLEN];

int Error (char *s)
{
  printf(s, lineNr+1, symBuf);
  errorCnt++;
  return 0;
}

int Number (char **p, int base)
{
  int v = 0, n = 0, c = **p - '0';
  char *q = symBuf;
  if(base != 10) *q++ = (*p)[-1];
  *q++ = c + '0';
  if(c > 9) {
    if(c < 17) Error("%d constant %s starts with invalid character\n");
    c -= 7;
  }
  *q = 0;
  if(c >= base) Error("%d digit in constant too large\n");
  while(c >= 0 && c < base) {
    n++; (*p)++; v = base*v + c;
    c = (*q++ = **p) - '0';
    if(c > 9) {
      if(c < 17) break;
      c -= 7;
    }
  }
  value[NSYMBOLS-1] = (v & 0xFFFF) | DEFINED; // KLUDGE: disguise result as symbol
  return NSYMBOLS-1;
}

int AlphaNum(char c)
{
  return (c >= '0' && c <= 'z' && (c >= 'a' || c <= '9' || c >= 'A' && c <= 'Z') || c == '_');
}

int ReadSymbol (char **p)
{
  char c, *q = symBuf;
  int hash1 = 0, hash2 = 0, index;
  White(p);
  c = **p;
  if(c == '$') { (*p)++; return Number(p, 16); }
  if(c == '%') { (*p)++; return Number(p, 2); }
  if(c == '0') { (*p)++; return Number(p, 8); }
  if(c >= '1' && c <= '9') return Number(p, 10);
  if(c < 'A' || c > 'z' || c < 'a' && c > 'Z') return 0;
  do {
    *q++ = c; (*p)++;
    hash1 = hash1 * (1 + 32 + 64 + 512) + c;
    hash2 = hash2 * (1 + 8 + 128) + c;
    if(q >= symBuf + SYMLEN) q--;
    c = **p;
  } while(AlphaNum(c));
  *q = 0;

  hash1 &= hashMask;
  hash1++; // make sure it is not 0
  if(hash1 > tabSize) hash1 -= tabSize;
  index = hash1;
  while(index) {
    if(!key[index]) { key[index] = hash2; break; } // add to table
    if(key[index] == hash2) break;
    index += hash1;
    if(index > tabSize) index -= tabSize;
  }
  if(pass2 && !(value[index] & DEFINED)) Error("%d undefined symbol '%s'\n"), value[index] |= DEFINED; // report only once
  return index;
}

char Operator (char **p)
{
  char c = **p;
  if(c == '+' || c == '-' || c == '|' || c == '&' || c == '^' || c == '*' || c == '/') (*p)++; else c = 0;
  return c;
}

int Term (char **p)
{
  int v, s;
  if(**p == '(') {
    (*p)++;
    v = Expression(p);
    if(**p != ')') Error("%d parentheses not closed\n"); else (*p)++;
    return v;
  }
  s = ReadSymbol(p);
  if(!s) { Error("%d missing operand\n"); return 0; }
  return value[s];
}

int Expression (char **p)
{
//int s; 
  int v, defined = DEFINED, high = 0, sign = 1;
  char c;
  White(p);
  if(**p == '>') high = 1, (*p)++; else
  if(**p == '<') high = -1, (*p)++;
  if(**p == '-') sign = -1, (*p)++; // unary minus
  White(p);
  v = Term(p);
  defined &= v;
  v *= sign;
  v &= 0xFFFF;
  White(p);
  while((c = Operator(p))) {
    int v2;
    White(p);
    v2 = Term(p);
    defined &= v2; v2 &= 0xFFFF;
    switch(c) {
      case '+': v += v2; break;
      case '-': v -= v2; break;
      case '*': v *= v2; break;
      case '/': v /= v2; break;
      case '&': v &= v2; break;
      case '|': v |= v2; break;
      case '^': v ^= v2; break;
    }
    v &= 0xFFFF;
    White(p);
  }
  if(high == 1) v >>= 8; else
  if(high == -1) v &= 0xFF;
  return v | defined;
}

int ReadMode (char **p)
{
  int v = 0, c, mode = 0, style = 0;
  White(p);
  c = **p;
  if(official) {
    if(c == '[') {
      (*p)++;
      if(**p == 'y' && (*p)[1] == ',') mode = 8, (*p) += 2;
      if(**p == 'x') {
        if((*p)[1] == ']') mode += (mode ? M_YINDEXED : M_INDEXED) | 5, (*p)++; else
        if((*p)[1] == '+' && (*p)[2] == '+' && mode) mode = M_INCREMENT | 29, (*p) += 3;
        else mode += 1;
      } else {
        v = Expression(p);
        mode = M_INDIRECT | 1;
        if(**p == ',' && (*p)[1] == 'x' && (*p)[2] == ']')
          (*p) += 2;
          if(!DX_EXTENSION && v & DEFINED && (v & 0xFF)) Error("%d indexed mode out of zero page not supported\n");
          else {
            mode = M_INDEXED | 5;
            v >>= 8;
          }
        }
      }
      if(**p != ']') return Error("%d closing ']' missing\n");
      (*p)++;
    } else if(c == 'i' && (*p)[1] == 'n' && !AlphaNum((*p)[2])) {
      (*p) += 2;
      mode = M_INPUT | 3;
    } else if(c == 'y' && (*p)[1] == ',') {
      (*p) += 2;
      v = Expression(p);
      mode = M_ABSOLUTE | 9;
    } else if(c == 'a' && (*p)[1] == 'c' && !AlphaNum((*p)[2])) {
      (*p) += 2;
      mode = M_ACCUMULATOR | 2;
    } else { //---------- not implemented yet
    Error("%d hardware enhancement (DX, Z) is not implemented yet!\n");
  }
  if(c == '[') {
    (*p)++;
    if(**p == 'A' && (*p)[1] == ']') mode = M_ACCUTARGET | 2, (*p)++; else
    if(**p == 'I' && (*p)[1] == 'N' && (*p)[2] == ']') mode = M_INTARGET | 3, (*p) += 2; else {
      if(**p == 'Y' && (*p)[1] == ':') mode = 8, (*p) += 2;
      if(**p == 'X') {
        if((*p)[1] == ']') mode += (mode ? M_YINDEXED : M_INDEXED) | 5, (*p)++; else
        if((*p)[1] == '+' && mode) mode = M_INCREMENT | 29, (*p) += 2;
        else mode += 1;
      } else {
        v = Expression(p);
        mode = M_INDIRECT | 1;
      }
    }
    White(p);
    if(**p != ']') return Error("%d closing ']' missing\n");
    (*p)++;
  } else if(c == '#') {
    (*p)++;
    v = Expression(p);
    mode = M_IMMEDIATE;
  } else if(c == 'I' && (*p)[1] == 'N' && !AlphaNum((*p)[2])) {
    (*p) += 2;
    mode = M_INPUT | 3;
  } else if(c == 'Y' && (*p)[1] == ':') {
    (*p) += 2;
    v = Expression(p);
    mode = M_ABSOLUTE | 9;
  } else if(c == 'A' && !AlphaNum((*p)[1])) {
    (*p)++;
    mode = M_ACCUMULATOR | 2;
  } else {
    v = Expression(p);
    mode = M_ZEROPAGE | 1;
    if(**p == ':' && (*p)[1] == 'X') {
      (*p) += 2;
      if(!DX_EXTENSION && v & DEFINED && (v & 0xFF)) Error("%d indexed mode out of zero page not supported\n");
      else {
        mode = M_INDEXED | 5;
        v >>= 8;
      }
    }
  }
  return v << 24 | mode;
}

int resultH, resultL, result, repeat;
int sync[] = { 0xC2, SAVA, 0, -1, 0x81, CYCLES, 0xC2, CYCLES, 0xE9, VIDEO, 0, -1, 0x01, SAVA };

int ParseOneLine (char **p)
{
  int indent = White(p);
  int s = ReadSymbol(p);
  result = 0;
  value[here] = DEFINED | curAddr[0]; // 'here' stands for current code address
  if(s) { // line starts with identifier (or number, which counts as predefined identifier)
    int v, adr, mode, n = value[s];
    if(!(n & MNEMONIC)) { // symbol is label or LHS assignment
      if(n & DEFINED && !pass2) return Error("%d %s doubly defined\n");
      White(p);
      if(**p == ':') { // label
        v = curAddr[segment]; (*p)++;
        indent = 2;
        yCont = -1; // this makes Y undefined
      } else if(**p == '=') { // assignment
        (*p)++; v = Expression(p);
        White(p);
        if(!EndOfLine(p)) return Error("%d error in expression\n");
        resultL = v & 0xFF; resultH = v >> 8 & 0xFF; result = 2;
      }
      v |= DEFINED;
      if(pass2 && v != n) return Error("%d %s phase error\n");
      value[s] = v;
      White(p);
      if(**p == '\n') return indent;
      s = ReadSymbol(p);
    } else indent = 1;
    // any preceeding label has now been processed, now do instruction
    n = value[s];
    if(!(n & (MNEMONIC | DIRECTIVE))) return Error("%d invalid mnemonic '%s'\n");
    adr = curAddr[segment];
    if(n & DIRECTIVE) {
      int x = 0, i, dir = (segment == 4 ? -1 : 1);
      switch(n & 255) {
        case 0: endFlag = 1; break;
        case 9: x++;                // vertical table space
        case 8: x++;                // frame page
        case 7: x++;                // zero page
        case 2: x++;                // data segment
        case 1: segment = x;        // code segment
          White(p);
          if(**p == '\n') break;
          v = Expression(p);
          if(!(v & DEFINED)) Error("%d undefined assembly address\n"); else
          if(segment >= 2 && v & 0xFF00) Error("%d address not in zero page\n");
          else curAddr[segment] = v & 0xFFFF;
          indent = 3;
          break;
        case 3: curAddr[segment] = (curAddr[segment] + 255) & ~255; break;
        case 4: curAddr[segment] += dir; break;
        case 5: White(p); curAddr[segment] += dir*Expression(p) & 0xFFFF; break;
        case 6:
          White(p); v = Expression(p);
          if(!(v & DEFINED)) Error("%d undefined repeat count\n"), v = 1;
          v = (v & 255) - 1;
          if(v > 0) repeat = v; break;
        case 10:
        case 11:
          White(p); v = Expression(p);
          n = (n & 1)*2;
          for(i=n; i<14-n; i++) printf("%d, %x\n",i,adr),objectCode[2*adr+i-n] = sync[i];
          objectCode[2*adr+3-n] = v - adr;
          objectCode[2*adr+11-n] = adr + 6 - n/2;
          curAddr[segment] += 7 - n;
          resultH = objectCode[2*adr]; resultL = objectCode[2*adr+1]; result = 14 - 2*n;
      }
      return indent;
    }
    if(n & M_MODE) { // must have addressing mode
      mode = ReadMode(p);
      if(!(mode & n)) return Error("%d addressing mode not allowed with this instruction\n");
      if((n & 0xFF) >= 224 && mode & M_ZEROPAGE) mode--;
      if((n & 0xFF) == 24 && mode & M_INCREMENT) mode -= 24;
      if(n & M_STORE) {
        mode &= ~3;
        White(p);
        if((n & 0xFF) == 192) { // MOV instruction needs second memory operand
          if(**p == ',') {
            (*p)++;
            White(p);
            v = ReadMode(p);
            if(!(v & (M_INDEXED | M_YINDEXED | M_INCREMENT))) Error("%d illegal destination on MOV instruction\n");
            mode = mode & 0xFF000000 | v-1 & 0xFFFFFF;
          } else Error("%d MOV needs second memory operand\n");
        }
      }
      if(**p == ',') {
        if(!(n & (M_REGISTER | M_ZEROPAGE | M_INCREMENT))) return Error("%d instruction does not allow second operand\n");
        (*p)++;
        if((*p)[1] == 'U' && (*p)[2] == 'T' && **p == 'O') (*p) += 2; // read away O and U
        if(AlphaNum((*p)[1])) return Error("%d invalid destination register\n");
        switch(**p) {
          case 'X': mode += 16; break;
          case 'Y': mode += 20; break;
          case 'T':
             if(n & M_STORE) Error("%d OUT cannot be loaded on store instructions\n");
             if(mode & M_ZEROPAGE) mode += 24;
             break;
          default: return Error("%d invalid destination register\n");
        }
        (*p)++;
      }
      n += mode;
      if(n & M_OPTIONAL && (mode & 0xFF0000FF) == yCont) return indent;
      if((n & 255) < 224 && (n & 0xFC) == 0x14) yCont = mode & 0xFF0000FF;
    }
    if(!segment) {
      int h = -1;
      if(!dx_ext && mode & M_INDEXED && mode & 0xFF<<24 && (mode & 0xFF<<24) != yCont) // const:X mode not equivalent to [X]
        h = n + 8, n = 0x14, yCont = mode & 0xFF<<24; // expand by first emitting LDY #const (spoiling Y)
      resultH = objectCode[2*adr] = n, resultL = objectCode[2*adr+1] = mode >> 24, result = 2 + 2*repeat;
      if(h != -1) objectCode[2*adr+2] = h, objectCode[2*adr+3] = 0, result += 2; // emit real const:X instruction
      curAddr[segment] += repeat;
      while(repeat > 0) objectCode[2*(adr+repeat)] = resultH, objectCode[2*(adr+repeat--)+1] = resultL;
    }
    curAddr[segment]++; repeat = 0;
    if(curAddr[2] & 0xFF00) Error("%d zero-page overflows\n");
  }
  White(p); if(**p != '\n') Error("%d garbage at end of line\n");
  return indent;
}

#define BUFLEN 1000
static char inbuf[BUFLEN];

void ListOneLine (int indent, int adr)
{
  char *p = inbuf;
  int i = indent ? 10 : 0;
  if(indent == 3) i += 10; else
  if(indent == 2) {
    do i--; while(*p++ != ':');
    if(i < 1) printf("%04X      ", adr);
    else if(result) printf("%04X %02X%02X ", adr, resultH, resultL);
    else printf("%04X      ", adr);
    p = inbuf;
    do printf("%c", *p++); while(*p != ':');
    printf(":"); p++;
    if(i < 1) {
      char *q = p;
      White(&q);
      if(*q == '\n') { printf("%s", p); return; }
      printf("\n");
      indent = 1; i = 20;
    }
    while(*p == ' ' || *p == '\t') p++;
  } else if(result) printf("%04X %02X%02X ", adr, resultH, resultL);
  else printf("%04X      ", adr);
  while(i-- > 0) printf(" ");
  printf("%s", p);
  while(result > 2) {
    adr++; result -= 2;
    printf("%04X %02X%02X\n", adr, objectCode[2*adr], objectCode[2*adr+1]);
  }
}
void Init ()
{
  int i, j=NSYMBOLS;
  char *p;
  tabSize = NSYMBOLS;
  do { // find prime < NSYMBOLS
    tabSize--;
    for(i=2; i<j; i++) {
      j = tabSize / i;
      if(i*j == tabSize) break;
    }
  } while(i < j);

  hashMask = ~(1<<31);
  while(hashMask > tabSize) hashMask >>= 1;

  // load all mnemonics in symbol table
  for(i=0; mnemonics[i].name; i++) {
    char *p = mnemonics[i].name;
    int s = ReadSymbol(&p);
    char buf[20];
    value[s] = mnemonics[i].opcode | MNEMONIC | DEFINED;
    strcpy(buf, mnemonics[i].name);
    p = buf;
    while(*p) *p++ += 'A' - 'a'; // capitalize
    p = buf;
    s = ReadSymbol(&p);
    value[s] = mnemonics[i].opcode | MNEMONIC | DEFINED;
  }
  p = "here";
  here = ReadSymbol(&p);
}

int ReadLine (FILE *f)
{
  char *p = inbuf;
  int c;
  while((c = getc(f)) != EOF) { *p++ = c; if(p > inbuf + BUFLEN - 2) p--; if(c == '\n') break; }
  *p = 0; lineNr++;
  return (c != EOF);
}

void Pass (char *source)
{
  FILE *f;
  f = fopen(source, "r");
  if(!f) fprintf(stderr, "Could not open source file %s\n", source), exit(0);
  errorCnt = endFlag = 0; lineNr = yCont = -1;
  curAddr[0] = curAddr[1] = curAddr[2] = curAddr[3] = 0; curAddr[4] = 0xFF;
  while(!endFlag && ReadLine(f)) {
    char *p = inbuf, indent;
    int adr = curAddr[segment];
    indent = ParseOneLine(&p);
    if(pass2 >= listing) ListOneLine(indent, adr);
  }
  fclose(f);
}

char *sourceFile = NULL, *objectFile = "ga.out";

void main (int argc, char **argv)
{
  while(argc > 1) {
    char *arg = argv[1];
    if(arg[0] == '-') {
      switch(arg[1]) {
        case 'l': // listing
          listing = 1;
          if(arg[2] == '1') listing = 0;
          break;
        case 'o':
          if(argc > 2) objectFile = argv[2], argc--, argv++;
          else fprintf(stderr, "missing object filename\n"), exit(0);
          break;
        case 'm':
          if(arg[2] == 'D' && arg[3] == 'X') dx_ext = 1; else
          if(arg[2] == 'Z') z_ext = 1; else
          fprintf(stderr, "unknown Gigatron extension %s\n", arg+1), exit(0);
          break;
      }
    } else if(!sourceFile) sourceFile = arg;
    else fprintf(stderr, "spurious argument %s\n", arg), exit(0);
    argc--; argv++;
  }
  if(!sourceFile) fprintf(stderr, "usage is: %s [-<option> ...] <sourcefile>\n", argv[0]),
    fprintf(stderr, "Options: -l           turn on listing (-l1 already in pass 1)\n"
                    "         -o <file>    object file (default is ga.out)\n"
                    "         -m<id>       use hardware enhancement (DX, Z)\n"), exit(0);
  Init();
  Pass(sourceFile);
  if(errorCnt) printf("pass 1 finished with %d errors\n", errorCnt), exit(0);
  pass2 = 1;
  Pass(sourceFile);
  if(!errorCnt) {
    FILE *f;
    printf("\nNo errors\n\n");
    f = fopen(objectFile, "w");
    if(!f) printf("could not open object file %s\n\n", objectFile), exit(0);
    fwrite(objectCode, 1, IMAGELEN, f);
    fclose(f);
  } else printf("\n*** %d errors in pass 2 ***\n\n", errorCnt);
}

/*
Notes to self

main modes #const, MEMORY, A, IN (opcode = 0, 1, 2, 3)
MEMORY has sub-modes addr, Y:addr, [X], [Y:X], [Y:X++] (opcode = 0, 8, 4, 12, 28-24 = 4)
second operand X, Y, OUT (opcode = 16, 20, 24)
mnemonic LD, AND, OR, XOR, ADD, SUB, ST, branches (opcode = 0, 32, 64, 96, 128, 160, 192, 224)
branches JMP, BGT, BLT, BNE, BEQ, BGE, BLE, BRA (opcode = 0, 4, 8, 12, 16, 20, 24, 28, all +224)
branch modes addr, [addr], [A], [IN] (opcode = 0, 1, 2, 3)
jmp has same modes as branch, prefixed by Y: (this could be implied)
Memonic values contain flag byte to indicate which modes and secondary operands are allowed

(c) 2017, H.G.Muller
*/